function C = join(A,colIndA,B,colIndB,mode)
% function C = join(A,colIndA,B,colIndB,mode)
% relational join of A and B, on the key columns colIndA and colIndB
% mode argument is optional (one of 'inner', 'l', 'r', 'lr')
% default mode: 'inner'
% the output columns are in the order [A B]
% for 'inner' and 'l' join, output columns are all columns from A, non-key columns from B
% for 'r' join, output columns are all non-key columns from A, all columns from B
% for 'lr' join, all columns from A and B are included
%
% Example:
% A = [ 3 4 5; 1 2 3; 6 7 8];
% B = [ 9 10 11; 3 4 55; 6 7 88];
% join(A,1:2,B,1:2)
%
% ans =
%
%     3     4     5    55
%     6     7     8    88
%
% join(A,1:2,B,1:2,'lr')
%
% ans =
%     1     2     3   NaN   NaN   NaN
%     3     4     5     3     4    55
%     6     7     8     6     7    88
%   NaN   NaN   NaN     9    10    11
%


% Stefan Schroedl
% 3/31/08

if nargin <5 
    mode = 'inner'; % default join mode
end

colIndA = unique(colIndA);
colIndB = unique(colIndB);

nA = size(A,1);
nB = size(B,1);

numColsA = size(A,2);
numColsB = size(B,2);

% non-key columns
colValA = setdiff(1:numColsA,colIndA);
colValB = setdiff(1:numColsB,colIndB);

% sort, attach original row number at end

AKeysSorted = sortrows([A(:,colIndA) (1:nA)']);
BKeysSorted = sortrows([B(:,colIndB) (1:nB)']);

% call mex function
[ia,ib] = join_merge(AKeysSorted(:,1:length(colIndA)),BKeysSorted(:,1:length(colIndB)),mode);

% construct output matrix
if (strcmp(mode,'inner'))
     C = [A(AKeysSorted(ia,end),:) B(BKeysSorted(ib,end), colValB)];
elseif (strcmp(mode,'l'))
     C = [A(AKeysSorted(ia,end),:) nan(length(ia),length(colValB))];
     idx = BKeysSorted(ib(ib>0),end);
     C(ia(ib>0), ((numColsA+1):end)) = B(idx, colValB);
elseif (strcmp(mode,'r'))
     C = [ nan(length(ia),length(colValA)) B(BKeysSorted(ib,end),:)];
     idx = AKeysSorted(ia(ia>0),end);
     C(ib(ia>0), 1:length(colValA)) = A(idx, colValA);
else
     % full outer mode 
     C = nan(length(ia),numColsA+numColsB);
     idx = AKeysSorted(ia(ia>0),end);
     C(ia>0, 1:numColsA) = A(idx, :);
     idx = BKeysSorted(ib(ib>0),end);
     C(ib>0, (numColsA+1):end) = B(idx,:);
end

