/*
  join_merge.cpp

  call: [index1,index2]=join_merge(mat1,mat2[, option])
  Assumption: 
   mat1 and mat2 have the same number of columns and 
   are sorted in lexicographic ascending order
   [option] specifies the type of join ('l', 'r', 'lr', or 'inner' (default))
  Ouput arguments have same size, and specify paired key row numbers in the 
  joined matrix (zero for unpaired)
  
  Stefan Schroedl
  3/13/08
 */

#include <iostream>
#include <math.h>
#include <stdlib.h>
#include <vector>
#include "mex.h"

using namespace std;

enum Mode { INNER = 0, L, R, LR };

Mode theMode;

extern void _main();

 char msg[1000];

void convertToDouble(double*& data,  mxArray ** ar)
{
  mxArray *mxdata;
  mexCallMATLAB(1, &mxdata, 1, ar, "double");
  data = mxGetPr(mxdata);
}

// compare two matrix rows lexicographically
// return -1 if first row smaller, 0 if equal, 1 if first row greater
inline int cmpRows( int rowA, int rowB, const double* a, const double* b, int numCols, int nA, int nB)
{
  double diff = 0;
  for (int i = 0; i < numCols; i++)
  {
    diff = a[rowA+nA*i] - b[rowB+nB*i]; // column-major storage
    if (diff != 0) break;
  }
  return ((diff < 0) ? -1 : ((diff == 0) ? 0 : 1));
}

void mexFunction(
		 int          nlhs,
		 mxArray      *plhs[],
		 int          nrhs,
		 const mxArray *prhs[]
		 )
{
  double      *keys1, *keys2;

  /* Check for proper number of arguments */

  if (nrhs < 2) {
    mexErrMsgTxt("JOIN_MERGE requires at least two input arguments.");
  };
  
  if (nlhs != 2) {
    mexErrMsgTxt("JOIN_MERGE requires two output arguments.");
  };
  
  if (nrhs == 3) {
    // user specified join mode
    const char* str = mxGetClassName(prhs[2]);
    if(strcmp(str,"char"))
    {
      mexErrMsgTxt("JOIN_MERGE third argument must be either 'inner', 'l', 'r', or 'lr'");      
    }

    int  length = mxGetN(prhs[2])+1;
    char*  strMode = (char*)mxCalloc(length, sizeof(char));
    mxGetString(prhs[2],strMode,length);
    if (!strcmp(strMode,"l")) {
      theMode = L;
    } else if  (!strcmp(strMode,"r")) {
      theMode = R;
    } else if (!strcmp(strMode,"lr")) {
      theMode = LR;
    } else if (!strcmp(strMode,"inner")) {
      theMode = INNER;
    } else {
      sprintf(msg, "JOIN_MERGE unknown join mode: %s\n", strMode);
      mexErrMsgTxt(msg);
    };
            
  }
 
  
  convertToDouble(keys1, (mxArray**)&prhs[0]);
 
  convertToDouble(keys2, (mxArray**)&prhs[1]);

   /* Find the dimensions of the data */
  int m1 = mxGetM(prhs[0]);
  int m2 = mxGetM(prhs[1]);
  int n1 = mxGetN(prhs[0]);
  int n2 = mxGetN(prhs[1]);

  if (n1 != n2) 
  {
    sprintf(msg, "JOIN_MERGE arguments must have same number of columns, %d, %d\n",m1,m2);
    mexErrMsgTxt(msg);
  };
 
  // output rows
  vector<int> rows1;
  vector<int> rows2;
  
  int cur1 = 0;
  int cur2 = 0;

  while (cur1 < m1 && cur2 < m2)
  {
    int cmp = cmpRows( cur1, cur2, keys1, keys2, n1, m1, m2);
    if (cmp < 0)
      {
	if (theMode&L)
	{
	  rows1.push_back(cur1);
	  rows2.push_back(-1);  // not pairable
	}
	cur1++;
      }
      else if (cmp > 0)
      {
	if (theMode&R)
	{
	  rows1.push_back(-1);  // not pairable
	  rows2.push_back(cur2);
	}
	cur2++;
      }
      else
      {
	// same key, now do join
	
        // first find duplicate keys
	int end1 = cur1+1;
	while (end1 < m1 && cmpRows(cur1, end1, keys1, keys1, n1, m1, m1) == 0)
	{
	  end1++;
	}
	int end2 = cur2+1;
	while (end2 < m2 && cmpRows(cur2, end2, keys2, keys2, n1, m2, m2) == 0)
	{
	  end2++;
	}

	// now add all cross product terms to output
	for (int i = cur1; i < end1; i++)
	  for (int j = cur2; j < end2; j++)
	  {
	    rows1.push_back(i);
	    rows2.push_back(j);
	  };
	cur1 = end1;
	cur2 = end2;
      }; 
    };

  // one of the arrays ran out

  if (cur1 < m1 && (theMode&L))
  {
    while (cur1 < m1)
    {
      rows1.push_back(cur1);
      rows2.push_back(-1);  // not pairable
      cur1++;
    }
  }
  if (cur2 < m2 && (theMode&R))
  {
    while (cur2 < m2)
    {
      rows1.push_back(-1); // not pairable
      rows2.push_back(cur2);
      cur2++;
    }
  }

  /* Create an mxArray for the output data */
  plhs[0] = mxCreateDoubleMatrix(rows1.size(), 1, mxREAL);
  plhs[1] = mxCreateDoubleMatrix(rows2.size(), 1, mxREAL);

  // copy output
  double* dataOut = mxGetPr(plhs[0]);
  for (int i = 0; i < rows1.size(); i++)
    dataOut[i] = rows1[i] + 1; // matlab indices start with 1
  
  dataOut = mxGetPr(plhs[1]);
  for (int i = 0; i < rows2.size(); i++)
  {
    dataOut[i] = rows2[i] + 1;
  };
}

